import numpy as np
from ..._trjwriterinterface import TRJWriterInterface


class POSCARWriter(TRJWriterInterface):
    fmt = "poscar"

    def write(self, atoms, coords, forces, velocities, *, box=None, **kwrgs):
        atoms = atoms.flatten()
        kind, numb = np.unique(atoms, return_counts=True)
        kind_line = "  ".join(kind) + "\n"
        numb_line = "  ".join(numb.astype(str)) + "\n"
        self.file.writelines(kind_line)
        scale = kwrgs.get("scale", 1.0)
        self.file.writelines(f" {scale:.16f}\n")
        box = box.astype(str)
        for r_box in box:
            self.file.writelines("  " + "  ".join(r_box) + "\n")
        self.file.writelines(" " + kind_line)
        self.file.writelines(" " + numb_line)
        self.file.writelines("Cartesian\n")
        for element in kind:
            for coord in coords[atoms == element]:
                self.file.writelines(f"{coord[0]:16f} {coord[1]:16f} {coord[2]:16f}\n")



import numpy as np
from numpy.typing import NDArray
from typing import Union
from .io import TRJOpener, TRJWriter


class Home(TRJOpener):
    def __init__(self, filename: str, *, fmt: str = "auto") -> None:
        super().__init__(filename, fmt=fmt)

    @property
    def atoms(self) -> NDArray[np.str_]:
        if not hasattr(self, "_atom_id"):
            self._atom_id = self.find_index(what=self.keys.get("atom"))
        return np.char.strip(self.data[:, self._atom_id])

    @property
    def coords(self) -> NDArray[np.float_]:
        if not hasattr(self, "_xyz_id"):
            self._xyz_id = self.find_index(what=["x", "y", "z"])
        return self.data[:, self._xyz_id].astype(float)

    @property
    def velocities(self) -> NDArray[np.float_]:
        if not hasattr(self, "_vxyz_id"):
            self._vxyz_id = self.find_index(what=["vx", "vy", "vz"])
        return self.data[:, self._vxyz_id].astype(float)

    @property
    def forces(self) -> NDArray[np.float_]:
        if not hasattr(self, "_fxyz_id"):
            self._fxyz_id = self.find_index(what=["fx", "fy", "fz"])
        return self.data[:, self._fxyz_id].astype(float)

    def find_index(self, what: Union[list[str], str]) -> NDArray[np.intp]:
        return np.where(np.isin(self.columns, what))[0]

    def find_atom(self, atom: str) -> NDArray[np.intp]:
        atom_indices = np.where(np.char.startswith(self.atoms, atom))[0]
        assert atom_indices.size, f"Atom({atom}) is not included in atom"
        return atom_indices

    def brew(self, what: Union[list[str], str] = None, atom: str = None) -> NDArray:
        atom_indices = self.find_atom(atom) if atom is not None else slice(None)
        col_indices = self.find_index(what) if what is not None else slice(None)
        if isinstance(atom_indices, np.ndarray) and isinstance(col_indices, np.ndarray):
            return self.data[np.ix_(atom_indices, col_indices)]
        return self.data[atom_indices, col_indices]

    def write(
        self,
        filename: str,
        start: int = 0,
        stop: int = None,
        step: int = 1,
        *,
        mode: str = "w",
        fmt: str = "auto",
        verbose: bool = True,
        **kwrgs,
    ) -> None:
        with TRJWriter(filename=filename, mode=mode, fmt=fmt) as f:
            for _ in self.frange(start=start, stop=stop, step=step, verbose=verbose):
                f.write(
                    atoms=self.atoms,
                    coords=self.coords,
                    forces=self.forces,
                    velocities=self.velocities,
                    box=self.box,
                    **kwrgs,
                )



from ..._trjwriterinterface import TRJWriterInterface


class XYZWriter(TRJWriterInterface):
    fmt = "xyz"

    def write(self, atoms, coords, forces, velocities, *, box=None, **kwrgs):
        natoms = len(atoms)
        self.file.writelines(f"\t{natoms}\n")
        self.file.writelines(f"\n")
        xyzlines = "\n".join(
            [f"{iatom[0]:>4s}\t{ixyz[0]:>16f}\t{ixyz[1]:>16f}\t{ixyz[2]:>16f}" for iatom, ixyz in zip(atoms, coords)]
        )
        self.file.writelines(xyzlines)
        self.file.writelines("\n")



from importlib import import_module
from pathlib import Path


HEAD = "atombrew.io.fmt"
for fmt_folder in Path(__file__).parent.glob("*"):
    for fmt_py in fmt_folder.glob("*.py"):
        if not "__" in fmt_folder.stem:
            import_module(f".{fmt_folder.stem}.{fmt_py.stem}", HEAD)
del import_module, Path



import numpy as np
from typing import TextIO
from ..._trjopenerinterface import TRJOpenerInterface


class POSCAROpener(TRJOpenerInterface):
    fmt = "poscar"
    _numb_additional_lines = 8

    def __init__(self, cls) -> None:
        super().__init__(cls)
        self.update_columns(np.array(["atom", "x", "y", "z"]))

    def _extract_snapshot(self, firstline: str, file: TextIO) -> list:
        SCALE = float(file.readline())
        box = np.loadtxt(file, max_rows=3, dtype=float) * SCALE
        self.update_box(box=box)
        kind_atoms = file.readline().split()
        numb_atoms = np.array(file.readline().split(), dtype=int)
        atoms = np.repeat(kind_atoms, numb_atoms)
        natoms = len(atoms)
        self.update_natoms(natoms=natoms)
        type_xyz = file.readline().strip()
        assert type_xyz == "Cartesian"
        xyz_lines = np.loadtxt(file, max_rows=natoms, dtype=str)
        return np.c_[atoms, xyz_lines]



import re
from collections import defaultdict
from ._color import get_color
from ._mass import get_mass, calc_molecularweights


def parse_formula(formula: str, *, is_summation: bool = True) -> dict[str, int]:
    # 정규 표현식을 사용하여 원소와 개수를 추출
    pattern = r"([A-Z][a-z]?)(\d*)"
    elements = re.findall(pattern, formula)
    # 결과를 저장할 딕셔너리 초기화
    result = defaultdict(int)
    # 추출된 원소와 개수를 딕셔너리에 추가
    if is_summation:
        for element, count in elements:
            count = int(count) if count else 1
            result[element] += count
    return dict(result)


__all__ = ["parse_formula" "get_color", "get_mass", "calc_molecularweights"]

del _mass, _color



from numpy import diag, array, tile
from atombrew import Home
from ._rdfinterface import RDFInterface


class ByHome(RDFInterface):
    _not_has_box_fmt = ["xyz"]

    def __init__(self, home: Home, what_a, what_b, box=None, r_max: float = None, resolution: int = 1000, dtype: str = ...):
        self.home = home.reset()
        self.what_a = self.home.find_atom(atom=what_a)
        self.what_b = self.home.find_atom(atom=what_b)
        self._is_box_none = box is None
        self.box = diag(self.home.box) if self._is_box_none else self._make_box(box=box)
        super().__init__(self.box, r_max, resolution, dtype)

    @property
    def n_a(self):
        if not hasattr(self, "_n_a"):
            self._n_a = len(self.what_a)
        return self._n_a

    @property
    def n_b(self):
        if not hasattr(self, "_n_b"):
            self._n_b = len(self.what_b)
        return self._n_b

    def run(self, start=0, stop=None, step=1):
        home = self.home
        what_a = self.what_a
        what_b = self.what_b
        for frame in self.home.frange(start, stop, step):
            coords = home.coords
            box = diag(home.box) if self._is_box_none else self.box[frame]
            self._update_hist(coords[what_a], coords[what_b], box)
        return self

    def _get_box(self, frame):
        pass
    
    def _make_box(self, box):
        box = array(box, dtype=float)
        if box.ndim == 3 and box.shape[1:] == (3,3):
            return array([diag(b) for b in box])
        elif box.ndim == 2:
            return box
        else:
            raise ValueError(f"Box should be in shape, (nframe, 3, 3), (nframe, 3)")


import numpy as np
from ..._trjwriterinterface import TRJWriterInterface


class LMPSWriter(TRJWriterInterface):
    fmt = "lmps"

    def write(self, atoms, coords, forces, velocities, *, box=None, **kwrgs):
        atoms = atoms.flatten()
        atomorder = kwrgs.get("atomorder", np.unique((atoms)))
        atomdict = {atom: i + 1 for i, atom in enumerate(atomorder)}
        file = self.file
        file.writelines(f"LMPS By AtomBrew\n\n")
        file.writelines(f"{len(atoms)} atoms\n")
        file.writelines(f"{len(atomdict)} atom types\n\n")
        for rb, a in zip(np.sum(box, axis=0), ["x", "y", "z"]):
            file.writelines(f"0 {rb:<16} {a}lo {a}hi\n")
        file.writelines("\n\nAtoms\n\n")
        if atoms[0].isalnum():
            atoms = [atomdict[atom] for atom in atoms]
        for i, (atom, coord) in enumerate(zip(atoms, coords)):
            file.writelines(f"{i+1} {atom} {coord[0]} {coord[1]} {coord[2]}\n")



import numpy as np
from typing import TextIO
from ..._trjopenerinterface import TRJOpenerInterface


WIDTH = [5, 5, 5, 5, 8, 8, 8, 8, 8, 8]
COLMUN = ["resnum", "resname", "atom", "id", "x", "y", "z", "vx", "vy", "vz"]


class GROOpener(TRJOpenerInterface):
    fmt = "gro"
    _numb_additional_lines = 3
    __is_update_column = False

    def __init__(self, cls) -> None:
        super().__init__(cls)
        self._width = WIDTH
        self.update_columns(COLMUN)

    def _extract_snapshot(self, firstline: str, file: TextIO) -> list:
        natom = int(file.readline().strip())
        data = np.genfromtxt(file, delimiter=self._width, dtype=str, max_rows=natom)
        box = np.diag([float(i) for i in file.readline().split()])
        self.update_box(box=box)
        self.update_natoms(natoms=natom)
        data = self.__check_data(data=data)
        return data

    def __check_data(self, data):
        if not self.__is_update_column:
            if "\n" in data[0]:
                self._width = self._width[:7]
                data = data[:, :7]
                self.update_columns(COLMUN[:7])
            self.__is_update_column = True
        return data



import numpy as np
from typing import TextIO
from ..._trjopenerinterface import TRJOpenerInterface


class LMPSOpener(TRJOpenerInterface):
    fmt = "lmps"
    _atom_keyword = "type"
    _numb_additional_lines = 12

    def __init__(self, cls) -> None:
        super().__init__(cls)
        self.update_columns(["id", self._atom_keyword, "x", "y", "z"])

    def _extract_snapshot(self, firstline: str, file: TextIO) -> list:
        next(file)
        # * Atom Line
        natoms = int(file.readline().split()[0])
        self.update_natoms(natoms=natoms)
        next(file)
        next(file)
        # * Box Line
        box = np.diag([sum([float(b) * (-1) ** (i + 1) for i, b in enumerate(file.readline().split()[:2])]) for _ in range(3)])
        self.update_box(box=box)
        next(file)
        next(file)
        next(file)
        next(file)
        return np.loadtxt(file, max_rows=natoms, dtype=str)



import numpy as np
from typing import TextIO
from ..._trjopenerinterface import TRJOpenerInterface


class XYZOpener(TRJOpenerInterface):
    fmt = "xyz"
    _numb_additional_lines = 2

    def __init__(self, cls) -> None:
        super().__init__(cls)
        self.update_columns(np.array(["atom", "x", "y", "z"]))

    def _extract_snapshot(self, firstline: str, file: TextIO) -> list:
        next(file)
        natoms = int(firstline.strip())
        self.update_natoms(natoms=natoms)

        return np.loadtxt(file, max_rows=natoms, dtype=str)



import numpy as np
from numpy.typing import ArrayLike


def apply_pbc(vec, box):
    box = np.array(box, dtype=float)
    vec = np.array(vec, dtype=float)
    half_box = box * 0.5
    return (vec + half_box) % box - half_box


def wrap(coords, box):
    box = np.array(box, dtype=float)
    coords = np.array(coords, dtype=float)
    return np.mod(coords, box)


def unwrap(coords, box):
    coords = np.array(coords, dtype=float)
    assert coords.ndim == 3, f"coords ndim != 3, your ndim: {coords.ndim}"
    box = np.array(box, dtype=float)
    assert box.ndim == 3, f"box ndim != 3, your ndim: {box.ndim}"
    unwrapped_xyz = np.copy(coords)
    for i in range(1, coords.shape[0]):
        ibox = np.diag(box[i])
        delta = coords[i] - coords[i - 1]
        delta -= np.round(delta / ibox) * ibox
        unwrapped_xyz[i] = unwrapped_xyz[i - 1] + delta
    return unwrapped_xyz



import numpy as np
from atombrew.space import calc_distance, apply_pbc
from abc import ABCMeta, abstractmethod


class RDFInterface(metaclass=ABCMeta):
    def __init__(
        self,
        box=None,
        r_max: float = None,
        resolution: int = 1000,
        dtype: str = float,
    ):
        self.nframe = 0
        self._dtype = dtype
        self.resolution = resolution
        self.r_max = np.max(box) * 0.5 if r_max is None else r_max
        self._histogram = np.zeros(self.resolution)
        self._rdf = np.zeros(self.resolution)

    def __repr__(self) -> str:
        return f"RDF{self.__class__.__name__}"

    @property
    def radii(self):
        if not hasattr(self, "_radii"):
            self._radii = np.linspace(0, self.r_max, num=self.resolution + 1)[:-1]
        return self._radii

    @property
    def result(self):
        if not hasattr(self, "_result"):
            self._result = self._cal_rdf()
        return self._result

    @property
    def cn(self):
        if not hasattr(self, "_cn"):
            self._cn = self._cal_cn()
        return self._cn

    @property
    @abstractmethod
    def n_a(self) -> int:
        pass

    @property
    @abstractmethod
    def n_b(self) -> int:
        pass

    @abstractmethod
    def run(self, start=0, end=None, step=1):
        pass

    def _update_hist(self, a_unit, b_unit, box_unit):
        vec = apply_pbc(a_unit[:, None, :] - b_unit[None, :, :], box=box_unit)
        distance = calc_distance(vec=vec)
        each_hist, _ = np.histogram(distance, bins=self.resolution, range=(0, self.r_max))
        self.nframe += 1
        self._histogram[1:] += each_hist[1:]
        self._rdf[1:] += each_hist[1:] / np.square(self.radii[1:]) * np.prod(box_unit)

    # Calculate the Density Function
    def _cal_rdf(self):
        dr = self.r_max / (self.resolution + 1)
        factor = 4.0 * np.pi * dr * self.nframe * self.n_a * self.n_b
        return self._rdf / factor

    # Function for get coordinate number
    def _cal_cn(self):
        return np.cumsum(self._histogram / (self.nframe * self.n_a))



import heapq
import itertools
import numpy as np
from scipy.spatial import cKDTree

__all__ = ["PeriodicCKDTree"]


def _gen_relevant_images(x, bounds, distance_upper_bound):
    # Map x onto the canonical unit cell, then produce the relevant
    # mirror images
    real_x = x - np.where(bounds > 0.0, np.floor(x / bounds) * bounds, 0.0)
    m = len(x)

    xs_to_try = [real_x]
    for i in range(m):
        if bounds[i] > 0.0:
            disp = np.zeros(m)
            disp[i] = bounds[i]

            if distance_upper_bound == np.inf:
                xs_to_try = list(itertools.chain.from_iterable((_ + disp, _, _ - disp) for _ in xs_to_try))
            else:
                extra_xs = []

                # Point near lower boundary, include image on upper side
                if abs(real_x[i]) < distance_upper_bound:
                    extra_xs.extend(_ + disp for _ in xs_to_try)

                # Point near upper boundary, include image on lower side
                if abs(bounds[i] - real_x[i]) < distance_upper_bound:
                    extra_xs.extend(_ - disp for _ in xs_to_try)

                xs_to_try.extend(extra_xs)

    return xs_to_try


class PeriodicCKDTree(cKDTree):
    """
    Cython kd-tree for quick nearest-neighbor lookup with periodic boundaries

    See scipy.spatial.ckdtree for details on kd-trees.

    Searches with periodic boundaries are implemented by mapping all
    initial data points to one canonical periodic image, building an
    ordinary kd-tree with these points, then querying this kd-tree multiple
    times, if necessary, with all the relevant periodic images of the
    query point.

    Note that to ensure that no two distinct images of the same point
    appear in the results, it is essential to restrict the maximum
    distance between a query point and a data point to half the smallest
    box dimension.
    """

    def __init__(self, data, bounds, leafsize=10):
        """Construct a kd-tree.

        Parameters
        ----------
        bounds : array_like, shape (k,)
            Size of the periodic box along each spatial dimension.  A
            negative or zero size for dimension k means that space is not
            periodic along k.
        data : array-like, shape (n,m)
            The n data points of dimension mto be indexed. This array is
            not copied unless this is necessary to produce a contiguous
            array of doubles, and so modifying this data will result in
            bogus results.
        leafsize : positive integer
            The number of points at which the algorithm switches over to
            brute-force.
        """

        # Map all points to canonical periodic image
        self.bounds = np.array(bounds)
        self.real_data = np.asarray(data)
        wrapped_data = self.real_data - np.where(
            self.bounds > 0.0, (np.floor(self.real_data / self.bounds) * self.bounds), 0.0
        )

        # Calculate maximum distance_upper_bound
        self.max_distance_upper_bound = np.min(np.where(self.bounds > 0, 0.5 * self.bounds, np.inf))

        # Set up underlying kd-tree
        super(PeriodicCKDTree, self).__init__(wrapped_data, leafsize)

    # Ideally, KDTree and cKDTree would expose identical query and __query
    # interfaces.  But they don't, and cKDTree.__query is also inaccessible
    # from Python.  We do our best here to cope.
    def __query(self, x, k=1, eps=0, p=2, distance_upper_bound=np.inf):
        # This is the internal query method, which guarantees that x
        # is a single point, not an array of points
        #
        # A slight complication: k could be "None", which means "return
        # all neighbors within the given distance_upper_bound".

        # Cap distance_upper_bound
        distance_upper_bound = np.min([distance_upper_bound, self.max_distance_upper_bound])

        # Run queries over all relevant images of x
        hits_list = []
        for real_x in _gen_relevant_images(x, self.bounds, distance_upper_bound):
            d, i = super(PeriodicCKDTree, self).query(real_x, k, eps, p, distance_upper_bound)
            if k > 1:
                hits_list.append(list(zip(d, i)))
            else:
                hits_list.append([(d, i)])

        # Now merge results
        if k > 1:
            return heapq.nsmallest(k, itertools.chain(*hits_list))
        elif k == 1:
            return [min(itertools.chain(*hits_list))]
        else:
            raise ValueError("Invalid k in periodic_kdtree._KDTree__query")

    def query(self, x, k=1, eps=0, p=2, distance_upper_bound=np.inf):
        """
        Query the kd-tree for nearest neighbors

        Parameters
        ----------
        x : array_like, last dimension self.m
            An array of points to query.
        k : integer
            The number of nearest neighbors to return.
        eps : non-negative float
            Return approximate nearest neighbors; the kth returned value
            is guaranteed to be no further than (1+eps) times the
            distance to the real k-th nearest neighbor.
        p : float, 1<=p<=infinity
            Which Minkowski p-norm to use.
            1 is the sum-of-absolute-values "Manhattan" distance
            2 is the usual Euclidean distance
            infinity is the maximum-coordinate-difference distance
        distance_upper_bound : nonnegative float
            Return only neighbors within this distance.  This is used to prune
            tree searches, so if you are doing a series of nearest-neighbor
            queries, it may help to supply the distance to the nearest neighbor
            of the most recent point.

        Returns
        -------
        d : array of floats
            The distances to the nearest neighbors.
            If x has shape tuple+(self.m,), then d has shape tuple+(k,).
            Missing neighbors are indicated with infinite distances.
        i : ndarray of ints
            The locations of the neighbors in self.data.
            If `x` has shape tuple+(self.m,), then `i` has shape tuple+(k,).
            Missing neighbors are indicated with self.n.

        """
        x = np.asarray(x)
        if np.shape(x)[-1] != self.m:
            raise ValueError("x must consist of vectors of length %d but " "has shape %s" % (self.m, np.shape(x)))
        if p < 1:
            raise ValueError("Only p-norms with 1<=p<=infinity permitted")
        retshape = np.shape(x)[:-1]
        if retshape != ():
            if k > 1:
                dd = np.empty(retshape + (k,), dtype=float)
                dd.fill(np.inf)
                ii = np.empty(retshape + (k,), dtype=int)
                ii.fill(self.n)
            elif k == 1:
                dd = np.empty(retshape, dtype=float)
                dd.fill(np.inf)
                ii = np.empty(retshape, dtype=int)
                ii.fill(self.n)
            else:
                raise ValueError(
                    "Requested %s nearest neighbors; acceptable "
                    "numbers are integers greater than or equal "
                    "to one, or None"
                )
            for c in np.ndindex(retshape):
                hits = self.__query(x[c], k=k, eps=eps, p=p, distance_upper_bound=distance_upper_bound)
                if k > 1:
                    for j in range(len(hits)):
                        dd[c + (j,)], ii[c + (j,)] = hits[j]
                elif k == 1:
                    if len(hits) > 0:
                        dd[c], ii[c] = hits[0]
                    else:
                        dd[c] = np.inf
                        ii[c] = self.n
            return dd, ii
        else:
            hits = self.__query(x, k=k, eps=eps, p=p, distance_upper_bound=distance_upper_bound)
            if k == 1:
                if len(hits) > 0:
                    return hits[0]
                else:
                    return np.inf, self.n
            elif k > 1:
                dd = np.empty(k, dtype=float)
                dd.fill(np.inf)
                ii = np.empty(k, dtype=int)
                ii.fill(self.n)
                for j in range(len(hits)):
                    dd[j], ii[j] = hits[j]
                return dd, ii
            else:
                raise ValueError(
                    "Requested %s nearest neighbors; acceptable "
                    "numbers are integers greater than or equal "
                    "to one, or None"
                )

    # Ideally, KDTree and cKDTree would expose identical __query_ball_point
    # interfaces.  But they don't, and cKDTree.__query_ball_point is also
    # inaccessible from Python.  We do our best here to cope.
    def __query_ball_point(self, x, r, p=2.0, eps=0):
        # This is the internal query method, which guarantees that x
        # is a single point, not an array of points

        # Cap r
        r = min(r, self.max_distance_upper_bound)

        # Run queries over all relevant images of x
        results = []
        for real_x in _gen_relevant_images(x, self.bounds, r):
            results.extend(super(PeriodicCKDTree, self).query_ball_point(real_x, r, p, eps))
        return results

    def query_ball_point(self, x, r, p=2.0, eps=0):
        """
        Find all points within distance r of point(s) x.

        Parameters
        ----------
        x : array_like, shape tuple + (self.m,)
            The point or points to search for neighbors of.
        r : positive float
            The radius of points to return.
        p : float, optional
            Which Minkowski p-norm to use.  Should be in the range [1, inf].
        eps : nonnegative float, optional
            Approximate search. Branches of the tree are not explored if their
            nearest points are further than ``r / (1 + eps)``, and branches are
            added in bulk if their furthest points are nearer than
            ``r * (1 + eps)``.

        Returns
        -------
        results : list or array of lists
            If `x` is a single point, returns a list of the indices of the
            neighbors of `x`. If `x` is an array of points, returns an object
            array of shape tuple containing lists of neighbors.

        Notes
        -----
        If you have many points whose neighbors you want to find, you may
        save substantial amounts of time by putting them in a
        PeriodicCKDTree and using query_ball_tree.
        """
        x = np.asarray(x).astype(float)
        if x.shape[-1] != self.m:
            raise ValueError("Searching for a %d-dimensional point in a " "%d-dimensional KDTree" % (x.shape[-1], self.m))
        if len(x.shape) == 1:
            return self.__query_ball_point(x, r, p, eps)
        else:
            retshape = x.shape[:-1]
            result = np.empty(retshape, dtype=object)
            for c in np.ndindex(retshape):
                result[c] = self.__query_ball_point(x[c], r, p, eps)
            return result

    def query_ball_tree(self, other, r, p=2.0, eps=0):
        raise NotImplementedError()

    def query_pairs(self, r, p=2.0, eps=0):
        raise NotImplementedError()

    def count_neighbors(self, other, r, p=2.0):
        raise NotImplementedError()

    def sparse_distance_matrix(self, other, max_distance, p=2.0):
        raise NotImplementedError()



from ._boundary import apply_pbc, wrap, unwrap
from ._calculate import calc_angle, calc_distance
from ._kdtree import PeriodicCKDTree

__all__ = ["apply_pbc", "wrap", "unwrap", "calc_angle", "calc_distance", "PeriodicCKDTree"]



import os
from ._trjwriterinterface import trj_writers


class TRJWriter(object):
    supporting_fmt = tuple(trj_writers.keys())

    def __init__(self, filename: str, mode="w", *, fmt: str = "auto") -> None:
        assert mode == "w+" or not os.path.isfile(filename), FileExistsError("Already File Exits")
        self._filename = filename
        self._mode = mode
        self.fmt = self._set_format(fmt=fmt)
        self._fmt_writer = trj_writers[self.fmt]

    def __enter__(self):
        self.file = open(self._filename, mode=self._mode)
        return self._fmt_writer(self.file)

    def __exit__(self, exc_type, exc_value, traceback):
        self.file.close()

    def _set_format(self, fmt: str):
        fmt = fmt.lower()
        if fmt == "auto":
            fmt = self._filename.split(".")[-1]
        assert fmt in self.supporting_fmt, f"Not Supporting Format({fmt}), We support {self.supporting_fmt}"
        return fmt



from typing import TextIO
from numpy.typing import NDArray
from abc import ABCMeta, abstractmethod


class TRJOpenerInterface(metaclass=ABCMeta):
    _atom_keyword = "atom"
    _numb_additional_lines = None

    def __init__(self, cls) -> None:
        self.cls = cls
        self.skip_headline_num = 0

    def __init_subclass__(cls) -> None:
        trj_openers[cls.fmt] = cls

    @property
    @abstractmethod
    def fmt(self) -> str:
        pass

    def extract_snapshot(self, file: TextIO) -> NDArray:
        firstline = file.readline()
        if not firstline:
            raise StopIteration()
        self.update_frame()
        return self._extract_snapshot(firstline=firstline, file=file)

    @abstractmethod
    def _extract_snapshot(self, firstline: str, file: TextIO):
        pass

    def update_frame(self):
        self.cls.frame += 1

    def update_box(self, box):
        self.cls._box = box

    def update_natoms(self, natoms):
        self.cls.natoms = natoms

    def update_columns(self, columns):
        self.cls._columns = columns


trj_openers: dict[str, type[TRJOpenerInterface]] = {}



from atombrew import Home
from ._main import subparsers

this_parser = subparsers.add_parser("trjconvert", help="Trajectory Convert")
this_parser.add_argument("-f", "--file", type=str, required=True, help="Token ID")
this_parser.add_argument("-o", "--out", type=str, required=True, help="Database ID")
this_parser.add_argument("-ofmt", type=str, default="auto", help="Open Format")
this_parser.add_argument("-wfmt", type=str, default="auto", help="Open Format")
this_parser.add_argument("-mode", type=str, default="w", help="Write Mode")
this_parser.add_argument("-start", type=int, default=0, help="Start Frame")
this_parser.add_argument("-end", type=int, default=None, help="End Frame")
this_parser.add_argument("-step", type=int, default=1, help="Frame Frequency")


def trjconvert(
    file,
    out,
    *,
    ofmt: str = "auto",
    wfmt: str = "auto",
    mode: str = "w",
    start=0,
    end=None,
    step=1,
    write_kwrgs: dict = None,
):

    home = Home(file, fmt=ofmt)
    home.write(out, fmt=wfmt, mode=mode, start=start, end=end, step=step, kwrgs=write_kwrgs)



import numpy as np
from tqdm import tqdm
from ._trjopenerinterface import trj_openers


class TRJOpener(object):
    supporting_fmt = tuple(trj_openers.keys())
    frame, natoms = 0, 0
    _box, _columns = [], []

    def __init__(self, filename: str, *, fmt: str = "auto") -> None:
        self._filename = filename
        self._fmt = self._set_format(fmt=fmt)
        self.reset()

    @property
    def fmt(self):
        return self._fmt

    @property
    def database(self):
        return self._database

    @property
    def data(self) -> np.ndarray:
        return self._data

    @property
    def box(self):
        return np.array(self._box)

    @property
    def columns(self):
        return np.array(self._columns)

    @property
    def keys(self):
        return {"atom": self._fmt_opener._atom_keyword}

    def __generate_db(self):
        with open(file=self._filename, mode="r") as file:
            for _ in range(self._fmt_opener.skip_headline_num):
                next(file)
            while True:
                try:
                    yield self._fmt_opener.extract_snapshot(file=file)
                except StopIteration:
                    break
                except Exception as e:
                    raise AssertionError(f"Unexpected error: {e}")

    def _set_format(self, fmt: str):
        fmt = str(fmt).lower()
        if fmt == "auto":
            fmt = self._filename.split(".")[-1]
        assert fmt in self.supporting_fmt, f"Not Supporting Format({fmt}), We support {self.supporting_fmt}"
        return fmt

    def load_db(self):
        self._database = self.__generate_db()

    def reset(self):
        self._fmt_opener = trj_openers[self._fmt](cls=self)
        self.frame, self._box = -1, []
        self.load_db()
        self.nextframe()
        return self

    def nextframe(self):
        self._data = next(self.database)

    def moveframe(self, frame: int):
        frame = int(frame)
        assert frame >= 0, f"frame should be positive integer"
        nlines_per_frame = self.natoms + self._fmt_opener._numb_additional_lines
        original_skip_headline_num = self._fmt_opener.skip_headline_num
        self._fmt_opener.skip_headline_num += nlines_per_frame * frame
        self.frame = frame - 1
        self._box = []
        self.load_db()
        self.nextframe()
        self._fmt_opener.skip_headline_num = original_skip_headline_num

    def frange(self, start: int = 0, stop: int = None, step: int = 1, *, verbose: bool = True, total: int = None):
        assert stop is None or start < stop, "start should be lower than stop"
        if stop is not None:
            total = stop - start
        if verbose:
            bar = tqdm(unit=" frame", total=total)
        self.moveframe(start)
        try:
            while self.frame != stop:
                if (self.frame - start) % step == 0:
                    yield self.frame
                self.nextframe()
                if verbose:
                    bar.update(1)
        except StopIteration:
            pass
        except Exception as e:
            raise AssertionError(f"Unexpected error: {e}")
        finally:
            self.reset()



import numpy as np
from typing import TextIO
from ..._trjopenerinterface import TRJOpenerInterface


class LAMMPSTRJOpener(TRJOpenerInterface):
    fmt = "lammpstrj"
    _atom_keyword = "type"
    _numb_additional_lines = 9

    def _extract_snapshot(self, firstline: str, file: TextIO) -> list:
        next(file)
        next(file)
        # * Atom Line
        natoms = int(file.readline().strip())
        self.update_natoms(natoms=natoms)
        # * Box Line
        next(file)
        box = np.diag([sum([float(b) * (-1) ** (i + 1) for i, b in enumerate(file.readline().split())]) for i in range(3)])
        self.update_box(box=box)
        # * Column Line
        columns = file.readline().split()[2:]
        if "element" in columns:
            self._atom_keyword = "element"
        self.update_columns(columns=columns)
        return np.loadtxt(file, max_rows=natoms, dtype=str)



import numpy as np


def calc_distance(vec, axis: int = -1, dtype: str = float):
    return np.sqrt(np.sum(np.square(vec), axis=axis)).astype(dtype)


def calc_angle(v1, v2):
    dot_product = np.sum(v1 * v2, axis=-1)
    norm_v1 = np.linalg.norm(v1, axis=-1)
    norm_v2 = np.linalg.norm(v2, axis=-1)
    return np.arccos(dot_product / (norm_v1 * norm_v2)) * 180.0 / np.pi



from ..._trjwriterinterface import TRJWriterInterface


class EXTXYZWriter(TRJWriterInterface):
    fmt = "extxyz"
    is_value_exist_dict = {
        "force": False,
    }

    def write(self, atoms, coords, forces, velocities, *, box=None, **kwrgs):
        file = self.file
        file.writelines(f"\t{len(atoms)}\n")
        file.writelines(self._make_infoline(box=box, isinforces=forces.size, isinvelocities=velocities.size))
        file.writelines(self._make_xyzline(atoms=atoms, coords=coords, forces=forces, velocities=velocities))

    def _make_infoline(self, box, isinforces: bool, isinvelocities: bool):
        assert box is not None, ValueError("box is None...")
        infoline = f'Lattice="' + " ".join(box.flatten().astype(str)) + '"'
        infoline += " Properties=species:S:1:pos:R:3"
        if isinforces:
            infoline += ":forces:R:3"
        if isinvelocities:
            infoline += ":velocities:R:3"
        infoline += ' pbc="True True True"\n'
        return infoline

    def _make_xyzline(self, atoms, coords, forces, velocities):
        xyzline = ""
        iterables = [atoms, coords]
        if forces.size:
            iterables.append(forces)
        if velocities.size:
            iterables.append(velocities)
        for atom, *properties in zip(*iterables):
            xyzline += f"{atom[0]:>4s}"
            for property in properties:
                xyzline += f" {property[0]:>16f} {property[1]:>16f} {property[2]:>16f}"
            xyzline += "\n"
        return xyzline



from pathlib import Path
from setuptools import setup, find_packages


version_dict = {}
with open(Path(__file__).parents[0] / "atombrew/__version.py") as this_v:
    exec(this_v.read(), version_dict)
version = version_dict["__version__"]
del version_dict


setup(
    name="atombrew",
    version=version,
    author="Knu",
    author_email="minu928@snu.ac.kr",
    url="https://github.com/minu928/atombrew",
    install_requires=["numpy>=1.22.3,<2.0.0", "tqdm>=4.66.4", "scipy>=1.11.4", "pandas>=2.1.4"],
    description="Package for post-processing the trajectory files of MD.",
    packages=find_packages(),
    python_requires=">=3.9",
    package_data={"": ["*"]},
    entry_points={
        "console_scripts": [
            "atb = atombrew.cli.__init__:parsing",
        ]
    },
    zip_safe=True,
)



import numpy as np
from typing import TextIO
from ..._trjopenerinterface import TRJOpenerInterface


PROPERTY_CANDIDATES = {
    "species": ["atom"],
    "atoms": ["atom"],
    "pos": ["x", "y", "z"],
    "positions": ["x", "y", "z"],
    "coords": ["x", "y", "z"],
    "forces": ["fx", "fy", "fz"],
    "vel": ["vx", "vy", "vz"],
    "velocities": ["vx", "vy", "vz"],
}
PROPERTY_KEYS = tuple(PROPERTY_CANDIDATES.keys())


class EXTXYZOpener(TRJOpenerInterface):
    fmt = "extxyz"
    _numb_additional_lines = 2
    is_update_columns = False

    def __init__(self, cls) -> None:
        super().__init__(cls)

    def _extract_snapshot(self, firstline: str, file: TextIO) -> list:
        natoms = int(firstline.strip())
        self.update_natoms(natoms=natoms)
        self._update_info(secondline=file.readline())
        return np.loadtxt(file, max_rows=natoms, dtype=str)

    def _update_info(self, secondline: str):
        # * Update Box
        box_str = secondline.split('Lattice="')[1].split('"')[0]
        box_str = np.array(box_str.split(), dtype=float).reshape(3, 3)
        self.update_box(box=box_str)
        # * Update Columns
        if not self.is_update_columns:
            columns = self._make_columns(secondline=secondline)
            self.update_columns(columns=columns)
            self.is_update_columns = True

    def _make_columns(self, secondline: str):
        if "Properties=" in secondline:
            properties = secondline.split("Properties=")[1].split()[0].split(":")[0::3]
            columns = []
            for property in properties:
                assert property in PROPERTY_KEYS, f"property:{property} not supported, {PROPERTY_KEYS}"
                this_column = PROPERTY_CANDIDATES.get(property)
                columns.extend(this_column)
            return columns
        return ["atom", "x", "y", "z"]



from scipy.constants import Avogadro
from .elements import parse_formula, calc_molecularweights
from typing import Dict, Union

# Constant for unit conversion (cubic angstrom to cubic cm)
ANGSTROM3_TO_CM3 = 1e-24


class Molecule:
    def __init__(self, formula: str) -> None:
        """
        Initialize a Molecule instance.

        Parameters
        ----------
        formula : str
            Chemical formula of the molecule.
        """
        self._formula = formula
        self._element_dict: Dict[str, int] = parse_formula(formula)
        self._mw: float = calc_molecularweights(element_count=self._element_dict)
        self._natoms: int = sum(self._element_dict.values())

    def __mul__(self, num: Union[int, float]) -> "Molecule":
        """
        Multiply the molecule by a number.

        Parameters
        ----------
        num : int or float
            Number to multiply the molecule by.

        Returns
        -------
        Molecule
            A new Molecule instance with multiplied formula.

        Raises
        ------
        TypeError
            If num is not a number.
        """
        if not isinstance(num, (int, float)):
            raise TypeError(f"Multiplication requires a number, got {type(num).__name__}")
        new_formula = "".join(f"{element}{count * num}" for element, count in self._element_dict.items())
        return Molecule(new_formula)

    def __add__(self, molecule: "Molecule") -> "Molecule":
        """
        Add two molecules together.

        Parameters
        ----------
        molecule : Molecule
            The molecule to add to this one.

        Returns
        -------
        Molecule
            A new Molecule instance representing the sum of the two molecules.

        Raises
        ------
        TypeError
            If the argument is not a Molecule instance.
        """
        if not isinstance(molecule, Molecule):
            raise TypeError(f"Addition requires an instance of Molecule, got {type(molecule).__name__}")
        return Molecule(self.formula + molecule.formula)

    def __repr__(self) -> str:
        """
        Return a string representation of the molecule.

        Returns
        -------
        str
            The chemical formula of the molecule.
        """
        return self.formula

    @property
    def mw(self) -> float:
        """
        Molecular weight of the molecule.

        Returns
        -------
        float
            Molecular weight in g/mol.
        """
        return self._mw

    @property
    def natoms(self) -> int:
        """
        Total number of atoms in the molecule.

        Returns
        -------
        int
            Number of atoms.
        """
        return self._natoms

    @property
    def formula(self) -> str:
        """
        Chemical formula of the molecule.

        Returns
        -------
        str
            The chemical formula.
        """
        return self._formula

    def calculate_volume(self, density: float) -> float:
        """
        Calculate volume of the molecule.

        Parameters
        ----------
        density : float
            Density in g/cm^3.

        Returns
        -------
        float
            Volume in Å^3.
        """
        return self.mw / (Avogadro * density * ANGSTROM3_TO_CM3)

    def calculate_density(self, volume: float) -> float:
        """
        Calculate density of the molecule.

        Parameters
        ----------
        volume : float
            Volume in Å^3.

        Returns
        -------
        float
            Density in g/cm^3.
        """
        return self.mw / (volume * Avogadro * ANGSTROM3_TO_CM3)

    def calculate_nmols(self, volume: float, density: float) -> float:
        """
        Calculate number of moles of the molecule.

        Parameters
        ----------
        volume : float
            Volume in Å^3.
        density : float
            Density in g/cm^3.

        Returns
        -------
        float
            Number of moles.
        """
        return density * volume * ANGSTROM3_TO_CM3 * Avogadro / self.mw



import numpy as np
from tqdm import trange
from ._rdfinterface import RDFInterface


class ByNumpy(RDFInterface):
    def __init__(self, a, b, box=None, r_max: float = None, resolution: int = 1000, dtype: str = ...):
        assert len(a) == len(b), f"Not equal frame of a({len(a)}) and b({len(b)})"
        self.a = np.array(a, dtype=float)
        self.b = np.array(b, dtype=float)
        self.box = self.__check_box(box=box)
        super().__init__(box, r_max, resolution, dtype)

    @property
    def n_a(self):
        if not hasattr(self, "_n_a"):
            self._n_a = self.a.shape[1]
        return self._n_a

    @property
    def n_b(self):
        if not hasattr(self, "_n_b"):
            self._n_b = self.b.shape[1]
        return self._n_b

    def run(self, start=0, stop=None, step=1):
        stop = len(self.a) if stop is None else stop
        for frame in trange(start, stop, step, unit="frame"):
            unit_a = self.a[frame, ...]
            unit_b = self.b[frame, ...]
            unit_box = self.box[frame, ...]
            self._update_hist(unit_a, unit_b, unit_box)
        return self

    def __check_box(self, box):
        assert box is not None, "Please Input Box"
        box = np.array(box, dtype=float)
        if box.ndim == 3 and box.shape[1:] == (3, 3):
            return np.array([np.diag(b) for b in box])
        elif box.ndim == 2:
            return box
        elif box.ndim == 1 and len(box) == 3:
            return np.tile(box, [len(self.a), 1])
        else:
            raise ValueError(f"Box should be in shape, (nframe, 3, 3), (nframe, 3), (3,)")



from ._main import parser
from ._trjconvert import trjconvert

func_dict = {"trjconvert": trjconvert}


def parsing():
    args = parser.parse_args().__dict__
    command = args.pop("command")
    return func_dict[command](**args)



import numpy as np
from tqdm import trange


class MSD(object):
    axis_dict = {"frame": 0, "natoms": 1, "dim": -1}

    def __init__(self, position, *, fft: bool = True, dtype: str = float):
        self.dtype = dtype
        self.fft = bool(fft)
        self.position = np.array(position, dtype=self.dtype)
        self.nframe = self.position.shape[0]

    def run(self, start: int = 0, stop: int = None, step: int = 1):
        if self.fft:
            self._result = self._calc_msd_with_fft()
        else:
            self._result = self._calc_msd_with_window()
        return self

    @property
    def result(self):
        if not hasattr(self, "_result"):
            self.run()
        return self._result

    def _calc_msd_with_window(self):
        msd_list = np.zeros(self.position.shape[:2])
        for frame in trange(1, self.nframe, unit="frame"):
            distance = self.__square_sum_position(self.position[frame:] - self.position[:-frame])
            msd_list[frame, :] = np.mean(distance, axis=self.axis_dict["frame"])
        return self.__mean_msd_list(msd_list=msd_list)

    def _calc_msd_with_fft(self):
        S_1 = self._calc_S1()
        S_2 = self._calc_S2()
        msd_list = np.subtract(S_1, 2.0 * S_2)
        return msd_list.mean(axis=self.axis_dict["natoms"])

    def _calc_S1(self):
        empty_matrix = np.zeros(self.position.shape[:2])
        D = np.square(self.position).sum(axis=self.axis_dict["dim"])
        D = np.append(D, empty_matrix, axis=self.axis_dict["frame"])
        Q = 2.0 * np.sum(D, axis=self.axis_dict["frame"])
        S_1 = empty_matrix
        for m in trange(self.nframe, unit="frame"):
            Q -= D[m - 1, :] + D[self.nframe - m, :]
            S_1[m, :] = Q / (self.nframe - m)
        return S_1

    def _calc_S2(self):
        X = np.fft.fft(self.position, n=2 * self.nframe, axis=self.axis_dict["frame"])
        dot_X = X * X.conjugate()
        x = np.fft.ifft(dot_X, axis=self.axis_dict["frame"])
        x = x[: self.nframe].real
        x = x.sum(axis=self.axis_dict["dim"])
        n = np.arange(self.nframe, 0, -1)
        return x / n[:, np.newaxis]



import numpy as np
from typing import TextIO
from ..._trjopenerinterface import TRJOpenerInterface


WIDTH = [6, 5, 1, 4, 1, 4, 1, 4, 1, 3, 8, 8, 8, 6, 6, 10, 2, 2]
COLMUN = np.array(
    [
        "pdbtype",
        "id",
        "_",
        "atom",
        "indicator",
        "resname",
        "chain",
        "resid",
        "rescode",
        "_",
        "x",
        "y",
        "z",
        "occupancy",
        "temp_factor",
        "seg_id",
        "element",
        "charge",
    ]
)


class PDBOpener(TRJOpenerInterface):
    fmt = "pdb"
    _numb_additional_lines = 3
    __natom = None
    __is_update_columns = False

    def __init__(self, cls) -> None:
        super().__init__(cls)
        self._width = WIDTH
        self.skip_headline_num = 2

    def _extract_snapshot(self, firstline: str, file: TextIO) -> list:
        box = np.diag([float(i) for i in file.readline().split()[1:4]])
        self.update_box(box=box)
        if self.__is_update_columns:
            data = np.genfromtxt(
                file,
                delimiter=self._width,
                dtype=str,
                max_rows=self.__natom,
                usecols=self._not_none_cols,
                autostrip=True,
            )
            next(file)
        else:
            start_indices = np.cumsum(np.concatenate(([0], WIDTH[:-1])))
            data = []
            while not (line := file.readline()).startswith("END"):
                idata = [line[start : start + width].strip() for start, width in zip(start_indices, WIDTH)]
                if not self.__is_update_columns:
                    self._not_none_cols = [i for i, cdata in enumerate(idata) if len(cdata)]
                    self.__is_update_columns = True
                    self.update_columns(COLMUN[self._not_none_cols])
                data.append(idata)
            data = np.array(data)[:, self._not_none_cols]
            self.__natom = len(data)
            self.update_natoms(natoms=self.__natom)
        return data



from typing import IO
from abc import ABCMeta, abstractmethod


class TRJWriterInterface(metaclass=ABCMeta):
    def __init__(self, file: IO) -> None:
        self.file = file

    def __init_subclass__(cls) -> None:
        trj_writers[cls.fmt] = cls

    @property
    @abstractmethod
    def fmt(self) -> str:
        pass

    @abstractmethod
    def write(self, atoms, coords, forces, velocities, *, box=None, **kwrgs):
        pass


trj_writers: dict[str, type[TRJWriterInterface]] = {}
